/*
 * Sprite.java
 *
 * Last modified: <2008/08/22 19:04:21 +0900> By SUMIDA
 */
package game;

import java.awt.Graphics;
import game.Surface;

import static java.lang.Math.*;
import static java.lang.System.*;
import static java.util.Arrays.*;
import static game.SystemInfo.*;
import static game.Real.*;
import static game.Logging.*;

public class Sprite {

    /* スプライト数 */
    public static final int SPRITE_MAX = 10000;

    /* オプション情報ビットアサイン */
    public static final int SPRITE_VISIBLE       = 0x00010000;
    public static final int SPRITE_USE_COLLISION = 0x00020000;
    public static final int SPRITE_USE_MOVE      = 0x00040000;
    public static final int SPRITE_USE_ANIME     = 0x00080000;
    public static final int SPRITE_USE_ALPHA     = 0x00100000;
    public static final int SPRITE_USE_LOWER     = 0x00200000;
    public static final int SPRITE_LOOP_ANIME    = 0x00400000;
    public static final int SPRITE_STOP_ANIME    = 0x00800000;
    public static final int SPRITE_FRAME         = 0x01000000;
    public static final int SPRITE_FRAME8        = 0x02000000;
    public static final int SPRITE_FRAME16       = 0x03000000;
    public static final int SPRITE_FRAME32       = 0x04000000;
    public static final int SPRITE_FRAME64       = 0x05000000;

    /* オプション情報マスク */
    public static final int SPRITE_MSK_FRAME     = 0xff000000;
    public static final int SPRITE_MSK_PRIORITY  = 0x000000ff;
    public static final int SPRITE_MSK_ALPHA     = 0x0000ff00;

    /* オプション情報シフト */
    public static final int SPRITE_SFT_FRAME     = 24;

    /* スプライト情報 */
    public static class Info implements Comparable<Info> {

        public int surface_id;     		/* サーフェスID       */
        public int frame_no;       		/* フレーム番号       */
        public int options;        		/* オプション情報     */
        public int x;              		/* 表示位置ｘ         */
        public int y;              		/* 表示位置ｙ         */
        public int move_interval;  		/* 移動間隔           */
        public int move_counter;   		/* 移動カウンタ       */
        public int dx;             		/* 移動量ｘ           */
        public int dy;             		/* 移動量ｙ           */
        public int frame_start;    		/* アニメ開始フレーム */
        public int frame_end;      		/* アニメ終了フレーム */
        public int frame_interval; 		/* フレーム遷移間隔   */
        public int frame_counter;  		/* フレームカウンタ   */
        public int collision_x0;   		/* 当たり判定矩形 x0  */
        public int collision_y0;   		/* 当たり判定矩形 y0  */
        public int collision_x1;   		/* 当たり判定矩形 x1  */
        public int collision_y1;   		/* 当たり判定矩形 y1  */

        public void init() {
            surface_id = 0;
            frame_no = 0;
            options = 0;
            x = 0;
            y = 0;
            move_interval = 0;
            move_counter  = 0;
            dx = 0;
            dy = 0;
            frame_start = 0;
            frame_end = 0;
            frame_interval = 0;
            frame_counter = 0;
            collision_x0 = 0;
            collision_y0 = 0;
            collision_x1 = 0;
            collision_y1 = 0;
        }

        @Override
        public int compareTo(Info o) {
            int pri0 = options & SPRITE_MSK_PRIORITY;
            int pri1 = o.options & SPRITE_MSK_PRIORITY;
            return pri0 - pri1;
        }
    }

    private static int sprite_count = 0;
    private static Info[] sprite_table = new Info[SPRITE_MAX];

    public static void init() {

        out.print("sprite_init ... ");
        sprite_count = 0;
        for (int i = 0; i < sprite_table.length; i++) {
            sprite_table[i] = null;
        }
        out.print("done.\n");
    }

    public static void clean() {
        sprite_count = 0;
    }

    public static void put(Info sprite) {

        if (sprite_count < sprite_table.length) {
            sprite_table[sprite_count++] = sprite;
        }
    }

    public static void rendering(Graphics screen) {

        if (sprite_count > 0) {
            int count = sprite_count;

            sort(sprite_table, 0, count);

            while (count-- > 0) {
                draw(screen, sprite_table[count]);
            }
        }
    }

    public static void draw(Graphics screen, Info sprite) {

        if ((sprite.options & SPRITE_VISIBLE) != 0) {

            int func_id = (sprite.options & SPRITE_MSK_FRAME) >> SPRITE_SFT_FRAME;

            int x = RealToInt(sprite.x);
            int y = RealToInt(sprite.y);

            if (Surface.isExists(sprite.surface_id) != true) {
                error("Surface is nothing. (" + sprite.surface_id + ")");
            } else {
                Surface.draw(func_id, screen, sprite.surface_id, sprite.frame_no, x, y);
            }
        } else {
            warn("Sprite is not visible.");
        }
    }

    public static void action(Info sprite) {

        if ((sprite.options & SPRITE_USE_MOVE) != 0) {

            if (++sprite.move_counter > sprite.move_interval) {
                sprite.move_counter = 0;
                sprite.x += sprite.dx;
                sprite.y += sprite.dy;
            }
        }

        if ((sprite.options & SPRITE_USE_ANIME) != 0
            && (sprite.options & SPRITE_STOP_ANIME) == 0) {

            if (++sprite.frame_counter > sprite.frame_interval) {
                sprite.frame_counter = 0;
                if (++sprite.frame_no > sprite.frame_end) {
                    if ((sprite.options & SPRITE_LOOP_ANIME) != 0) {
                        sprite.frame_no = sprite.frame_start;
                    } else {
                        sprite.frame_no = sprite.frame_end;
                        sprite.options |= SPRITE_STOP_ANIME;
                    }
                }
            }
        }

        if ((sprite.options & SPRITE_USE_LOWER) != 0) {

            int pri = sprite.options & SPRITE_MSK_PRIORITY;

            if (pri < SPRITE_MSK_PRIORITY) {
                sprite.options &= ~SPRITE_MSK_PRIORITY;
                sprite.options |= pri + 1;
            }
        }
    }

    public static boolean collisionCheck(Info s0, Info s1) {

        boolean result = false;

        if ((s0.options & SPRITE_USE_COLLISION) != 0
            && (s1.options & SPRITE_USE_COLLISION) != 0) {

            int x0 = s1.x + s1.collision_x0 - s0.x;
            int y0 = s1.y + s1.collision_y0 - s0.y;
            int x1 = s1.x + s1.collision_x1 - s0.x;
            int y1 = s1.y + s1.collision_y1 - s0.y;

            if (x0 < s0.collision_x1 && x1 > s0.collision_x0
                && y0 < s0.collision_y1 && y1 > s0.collision_y0) {
                result = true;
            }
        }

        return result;
    }

    public static double getAngle(Info s0, Info s1) {

        Surface.Info surface0 = Surface.getInfo(s0.surface_id);
        Surface.Info surface1 = Surface.getInfo(s1.surface_id);

        double dx = (s1.x + (IntToReal(surface1.frame_width) >> 1)
                     - s0.x - (IntToReal(surface0.frame_width) >> 1)) / 65536.0;
        double dy = (s1.y + (IntToReal(surface1.frame_height) >> 1)
                     - s0.y - (IntToReal(surface0.frame_height) >> 1)) / 65536.0;

        return angle(dx, dy);
    }

    public static void setAngle(Info s, double angle, int speed) {

        double r = angle * PI / 180;

        s.dx = (int)(sin(r) * speed);
        s.dy = (int)(cos(r) * speed);

        if (angle >= 180) {
            s.dx = -s.dx;
            s.dy = -s.dy;
        }
    }

    public static void fire(Info sprite, Info parent, int ox, int oy) {

        Surface.Info surface_this = Surface.getInfo(sprite.surface_id);
        Surface.Info surface_parent = Surface.getInfo(parent.surface_id);

        sprite.x = parent.x + ox
            + (IntToReal(surface_parent.frame_width - surface_this.frame_width) >> 1);
        sprite.y = parent.y + oy
            + (IntToReal(surface_parent.frame_height - surface_this.frame_height) >> 1);
    }

    public static void homing(Info sprite, Info target, int speed) {

        Surface.Info surface_this = Surface.getInfo(sprite.surface_id);
        Surface.Info surface_target = Surface.getInfo(target.surface_id);

        double dx = (target.x + (IntToReal(surface_target.frame_width) >> 1)
                     - sprite.x - (IntToReal(surface_this.frame_width) >> 1)) / 65536.0;
        double dy = (target.y + (IntToReal(surface_target.frame_height) >> 1)
                     - sprite.y - (IntToReal(surface_this.frame_height) >> 1)) / 65536.0;
        double d = sqrt(dx * dx + dy * dy);

        if (d != 0) {
            double spd = speed / 65536.0;
            sprite.dx = (int)((dx / d * spd) * 65536);
            sprite.dy = (int)((dy / d * spd) * 65536);
        } else {
            sprite.dx = 0;
            sprite.dy = speed;
        }
    }
}
