/*
 * Surface.java
 *
 * Last modified: <2008/08/19 18:56:20 +0900> By SUMIDA
 */
package game;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Toolkit;
import java.net.URL;

import static java.lang.System.*;

public class Surface {

    public static class Info {
        public Image surface;
        public int frame_width;
        public int frame_height;
    }

    public static final int SURFACES_MAX = 256;
    public static final byte FUNCID_SURFACE_DRAW     = 1;
    public static final byte FUNCID_SURFACE_DRAW_8   = 2;
    public static final byte FUNCID_SURFACE_DRAW_16  = 3;
    public static final byte FUNCID_SURFACE_DRAW_32  = 4;
    public static final byte FUNCID_SURFACE_DRAW_64  = 5;

    public static Info[] surfaces = new Info[SURFACES_MAX];

    private static Component canvas;

    public static void init(Component comp) {

        out.print("surface_init ... ");
        canvas = comp;
        for (int i = 0 ; i < surfaces.length; i++) {
            surfaces[i] = null;
        }
        out.print("done.\n");
    }

    public static Info load(final String name, final int id) {

        out.print("surface_load(" + name + ") ... ");

        Info info = new Info();
        surfaces[id] = info;

        try {
            MediaTracker tracker = new MediaTracker(canvas);
            URL url = Surface.class.getClassLoader().getResource(name);
            Image image = Toolkit.getDefaultToolkit().getImage(url);
            tracker.addImage(image, 0);

            // イメージ読み込み完了まで待機
            try {
                tracker.waitForID(0);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }

            /* 今回のサーフェスを登録 */
            info.surface      = image;
            info.frame_width  = image.getWidth(null);
            info.frame_height = image.getHeight(null);

            out.print("done.\n");

        } catch (Exception e) {
            e.printStackTrace();
        }

        return info;
    }

    public static void trash(int id) {

        Info info = surfaces[id];
        if (info != null) {
            info.surface = null;
            surfaces[id] = null;
        }
    }

    public static void free() {

        out.print("surface_free ... ");

        for (int id = 0; id < surfaces.length; id++) {
            trash(id);
        }

        out.print("done.\n");
    }

    public static void setFrame(int id, int width, int height) {

        Info info = surfaces[id];

        info.frame_width  = width;
        info.frame_height = height;
    }

    public static void draw(int funcid, Graphics screen, int id, int frame, int x, int y) {
        switch (funcid) {
        case FUNCID_SURFACE_DRAW:
            drawFrame(screen, id, frame, x, y);
            break;
        case FUNCID_SURFACE_DRAW_8:
            drawFrame8(screen, id, frame, x, y);
            break;
        case FUNCID_SURFACE_DRAW_16:
            drawFrame16(screen, id, frame, x, y);
            break;
        case FUNCID_SURFACE_DRAW_32:
            drawFrame32(screen, id, frame, x, y);
            break;
        case FUNCID_SURFACE_DRAW_64:
            drawFrame64(screen, id, frame, x, y);
            break;
        default:
            draw(screen, id, x, y);
            break;
        }
    }

    public static void draw(Graphics screen, int id, int x, int y) {

        Info info = surfaces[id];

        //out.print("surface_draw(" + id + ") ... ");

        screen.drawImage(info.surface,
                         x,
                         y,
                         x + info.frame_width,
                         y + info.frame_height,
                         0,
                         0,
                         info.frame_width,
                         info.frame_height,
                         null);

        //out.print("done.\n");
    }

    public static void drawFrame(Graphics screen, int id, int frame, int x, int y) {

        Info info = surfaces[id];

        int w = info.surface.getWidth(null) / info.frame_width;
        int sx = (frame % w) * info.frame_width;
        int sy = (frame / w) * info.frame_height;

        screen.drawImage(info.surface,
                         x,
                         y,
                         x + info.frame_width,
                         y + info.frame_height,
                         sx,
                         sy,
                         sx + info.frame_width,
                         sy + info.frame_height,
                         null);
    }

    public static void drawFrame8(Graphics screen, int id, int frame, int x, int y) {

        Info info = surfaces[id];

        int w = info.surface.getWidth(null) >> 3;
        int sx = (frame % w) << 3;
        int sy = (frame / w) << 3;

        screen.drawImage(info.surface,
                         x,
                         y,
                         x + 8,
                         y + 8,
                         sx,
                         sy,
                         sx + 8,
                         sy + 8,
                         null);
    }

    public static void drawFrame16(Graphics screen, int id, int frame, int x, int y) {

        Info info = surfaces[id];

        int w = info.surface.getWidth(null) >> 4;
        int sx = (frame % w) << 4;
        int sy = (frame / w) << 4;

        screen.drawImage(info.surface,
                         x,
                         y,
                         x + 16,
                         y + 16,
                         sx,
                         sy,
                         sx + 16,
                         sy + 16,
                         null);
    }

    public static void drawFrame32(Graphics screen, int id, int frame, int x, int y) {

        Info info = surfaces[id];

        int w = info.surface.getWidth(null) >> 5;
        int sx = (frame % w) << 5;
        int sy = (frame / w) << 5;

        screen.drawImage(info.surface,
                         x,
                         y,
                         x + 32,
                         y + 32,
                         sx,
                         sy,
                         sx + 32,
                         sy + 32,
                         null);
    }

    public static void drawFrame64(Graphics screen, int id, int frame, int x, int y) {

        Info info = surfaces[id];

        int w = info.surface.getWidth(null) >> 6;
        int sx = (frame % w) << 6;
        int sy = (frame / w) << 6;

        screen.drawImage(info.surface,
                         x,
                         y,
                         x + 64,
                         y + 64,
                         sx,
                         sy,
                         sx + 64,
                         sy + 64,
                         null);
    }

    public static Info getInfo(int id) {

        return surfaces[id];
    }

    public static boolean isExists(int id) {

        boolean result = false;
        Info info = surfaces[id];

        if (id >= 0 && id < surfaces.length
            && info != null && info.surface != null) result = true;

        return result;
    }
}
